from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator

from dep_logic.markers.any import AnyMarker
from dep_logic.markers.base import BaseMarker, EvaluationContext
from dep_logic.markers.empty import EmptyMarker
from dep_logic.markers.single import MarkerExpression, SingleMarker
from dep_logic.utils import DATACLASS_ARGS, flatten_items, intersection, union


@dataclass(init=False, frozen=True, unsafe_hash=True, **DATACLASS_ARGS)
class MultiMarker(BaseMarker):
    markers: tuple[BaseMarker, ...]

    def __init__(self, *markers: BaseMarker) -> None:
        object.__setattr__(self, "markers", tuple(flatten_items(markers, MultiMarker)))

    def __iter__(self) -> Iterator[BaseMarker]:
        return iter(self.markers)

    @property
    def complexity(self) -> tuple[int, ...]:
        return tuple(sum(c) for c in zip(*(m.complexity for m in self.markers)))

    @classmethod
    def of(cls, *markers: BaseMarker) -> BaseMarker:
        from dep_logic.markers.union import MarkerUnion

        new_markers = flatten_items(markers, MultiMarker)
        old_markers: list[BaseMarker] = []

        while old_markers != new_markers:
            old_markers = new_markers
            new_markers = []
            for marker in old_markers:
                if marker in new_markers:
                    continue

                if marker.is_any():
                    continue

                intersected = False
                for i, mark in enumerate(new_markers):
                    # If we have a SingleMarker then with any luck after intersection
                    # it'll become another SingleMarker.
                    if isinstance(mark, SingleMarker):
                        new_marker = mark & marker
                        if new_marker.is_empty():
                            return EmptyMarker()

                        if isinstance(new_marker, SingleMarker):
                            new_markers[i] = new_marker
                            intersected = True
                            break

                    # If we have a MarkerUnion then we can look for the simplifications
                    # implemented in intersect_simplify().
                    elif isinstance(mark, MarkerUnion):
                        intersection = mark.intersect_simplify(marker)
                        if intersection is not None:
                            new_markers[i] = intersection
                            intersected = True
                            break

                if intersected:
                    # flatten again because intersect_simplify may return a multi
                    new_markers = flatten_items(new_markers, MultiMarker)
                    continue

                new_markers.append(marker)

        if any(m.is_empty() for m in new_markers):
            return EmptyMarker()

        if not new_markers:
            return AnyMarker()

        if len(new_markers) == 1:
            return new_markers[0]

        return MultiMarker(*new_markers)

    def __and__(self, other: BaseMarker) -> BaseMarker:
        return intersection(self, other)

    def __or__(self, other: BaseMarker) -> BaseMarker:
        return union(self, other)

    __rand__ = __and__
    __ror__ = __or__

    def union_simplify(self, other: BaseMarker) -> BaseMarker | None:
        """
        Finds a couple of easy simplifications for union on MultiMarkers:

            - union with any marker that appears as part of the multi is just that
              marker

            - union between two multimarkers where one is contained by the other is just
              the larger of the two

            - union between two multimarkers where there are some common markers
              and the union of unique markers is a single marker
        """
        if other in self.markers:
            return other

        if isinstance(other, MultiMarker):
            our_markers = set(self.markers)
            their_markers = set(other.markers)

            if our_markers.issubset(their_markers):
                return self

            if their_markers.issubset(our_markers):
                return other

            shared_markers = our_markers.intersection(their_markers)
            if not shared_markers:
                return None

            unique_markers = our_markers - their_markers
            other_unique_markers = their_markers - our_markers
            unique_union = MultiMarker(*unique_markers) | (
                MultiMarker(*other_unique_markers)
            )
            if isinstance(unique_union, (SingleMarker, AnyMarker)):
                # Use list instead of set for deterministic order.
                common_markers = [
                    marker for marker in self.markers if marker in shared_markers
                ]
                return unique_union & MultiMarker(*common_markers)

        return None

    def evaluate(
        self,
        environment: dict[str, str | set[str]] | None = None,
        context: EvaluationContext = "metadata",
    ) -> bool:
        return all(m.evaluate(environment, context) for m in self.markers)

    def without_extras(self) -> BaseMarker:
        return self.exclude("extra")

    def exclude(self, marker_name: str) -> BaseMarker:
        new_markers = []

        for m in self.markers:
            if isinstance(m, SingleMarker) and m.name == marker_name:
                # The marker is not relevant since it must be excluded
                continue

            marker = m.exclude(marker_name)

            if not marker.is_empty():
                new_markers.append(marker)

        return self.of(*new_markers)

    def only(self, *marker_names: str) -> BaseMarker:
        return self.of(*(m.only(*marker_names) for m in self.markers))

    def __str__(self) -> str:
        elements = []
        for m in self.markers:
            if isinstance(m, (MarkerExpression, MultiMarker)):
                elements.append(str(m))
            else:
                elements.append(f"({m})")

        return " and ".join(elements)



import pytest

from dep_logic.tags import os
from dep_logic.tags.platform import Arch, Platform


def test_platform_tags_manylinux():
    tags = Platform(os.Manylinux(2, 20), Arch.X86_64).compatible_tags
    assert tags == [
        "manylinux_2_20_x86_64",
        "manylinux_2_19_x86_64",
        "manylinux_2_18_x86_64",
        "manylinux_2_17_x86_64",
        "manylinux2014_x86_64",
        "manylinux_2_16_x86_64",
        "manylinux_2_15_x86_64",
        "manylinux_2_14_x86_64",
        "manylinux_2_13_x86_64",
        "manylinux_2_12_x86_64",
        "manylinux2010_x86_64",
        "manylinux_2_11_x86_64",
        "manylinux_2_10_x86_64",
        "manylinux_2_9_x86_64",
        "manylinux_2_8_x86_64",
        "manylinux_2_7_x86_64",
        "manylinux_2_6_x86_64",
        "manylinux_2_5_x86_64",
        "manylinux1_x86_64",
        "linux_x86_64",
    ]


def test_platform_tags_macos():
    tags = Platform(os.Macos(21, 6), Arch.X86_64).compatible_tags
    assert tags == [
        "macosx_21_0_x86_64",
        "macosx_21_0_intel",
        "macosx_21_0_fat64",
        "macosx_21_0_fat32",
        "macosx_21_0_universal2",
        "macosx_21_0_universal",
        "macosx_20_0_x86_64",
        "macosx_20_0_intel",
        "macosx_20_0_fat64",
        "macosx_20_0_fat32",
        "macosx_20_0_universal2",
        "macosx_20_0_universal",
        "macosx_19_0_x86_64",
        "macosx_19_0_intel",
        "macosx_19_0_fat64",
        "macosx_19_0_fat32",
        "macosx_19_0_universal2",
        "macosx_19_0_universal",
        "macosx_18_0_x86_64",
        "macosx_18_0_intel",
        "macosx_18_0_fat64",
        "macosx_18_0_fat32",
        "macosx_18_0_universal2",
        "macosx_18_0_universal",
        "macosx_17_0_x86_64",
        "macosx_17_0_intel",
        "macosx_17_0_fat64",
        "macosx_17_0_fat32",
        "macosx_17_0_universal2",
        "macosx_17_0_universal",
        "macosx_16_0_x86_64",
        "macosx_16_0_intel",
        "macosx_16_0_fat64",
        "macosx_16_0_fat32",
        "macosx_16_0_universal2",
        "macosx_16_0_universal",
        "macosx_15_0_x86_64",
        "macosx_15_0_intel",
        "macosx_15_0_fat64",
        "macosx_15_0_fat32",
        "macosx_15_0_universal2",
        "macosx_15_0_universal",
        "macosx_14_0_x86_64",
        "macosx_14_0_intel",
        "macosx_14_0_fat64",
        "macosx_14_0_fat32",
        "macosx_14_0_universal2",
        "macosx_14_0_universal",
        "macosx_13_0_x86_64",
        "macosx_13_0_intel",
        "macosx_13_0_fat64",
        "macosx_13_0_fat32",
        "macosx_13_0_universal2",
        "macosx_13_0_universal",
        "macosx_12_0_x86_64",
        "macosx_12_0_intel",
        "macosx_12_0_fat64",
        "macosx_12_0_fat32",
        "macosx_12_0_universal2",
        "macosx_12_0_universal",
        "macosx_11_0_x86_64",
        "macosx_11_0_intel",
        "macosx_11_0_fat64",
        "macosx_11_0_fat32",
        "macosx_11_0_universal2",
        "macosx_11_0_universal",
        "macosx_10_16_x86_64",
        "macosx_10_16_intel",
        "macosx_10_16_fat64",
        "macosx_10_16_fat32",
        "macosx_10_16_universal2",
        "macosx_10_16_universal",
        "macosx_10_15_x86_64",
        "macosx_10_15_intel",
        "macosx_10_15_fat64",
        "macosx_10_15_fat32",
        "macosx_10_15_universal2",
        "macosx_10_15_universal",
        "macosx_10_14_x86_64",
        "macosx_10_14_intel",
        "macosx_10_14_fat64",
        "macosx_10_14_fat32",
        "macosx_10_14_universal2",
        "macosx_10_14_universal",
        "macosx_10_13_x86_64",
        "macosx_10_13_intel",
        "macosx_10_13_fat64",
        "macosx_10_13_fat32",
        "macosx_10_13_universal2",
        "macosx_10_13_universal",
        "macosx_10_12_x86_64",
        "macosx_10_12_intel",
        "macosx_10_12_fat64",
        "macosx_10_12_fat32",
        "macosx_10_12_universal2",
        "macosx_10_12_universal",
        "macosx_10_11_x86_64",
        "macosx_10_11_intel",
        "macosx_10_11_fat64",
        "macosx_10_11_fat32",
        "macosx_10_11_universal2",
        "macosx_10_11_universal",
        "macosx_10_10_x86_64",
        "macosx_10_10_intel",
        "macosx_10_10_fat64",
        "macosx_10_10_fat32",
        "macosx_10_10_universal2",
        "macosx_10_10_universal",
        "macosx_10_9_x86_64",
        "macosx_10_9_intel",
        "macosx_10_9_fat64",
        "macosx_10_9_fat32",
        "macosx_10_9_universal2",
        "macosx_10_9_universal",
        "macosx_10_8_x86_64",
        "macosx_10_8_intel",
        "macosx_10_8_fat64",
        "macosx_10_8_fat32",
        "macosx_10_8_universal2",
        "macosx_10_8_universal",
        "macosx_10_7_x86_64",
        "macosx_10_7_intel",
        "macosx_10_7_fat64",
        "macosx_10_7_fat32",
        "macosx_10_7_universal2",
        "macosx_10_7_universal",
        "macosx_10_6_x86_64",
        "macosx_10_6_intel",
        "macosx_10_6_fat64",
        "macosx_10_6_fat32",
        "macosx_10_6_universal2",
        "macosx_10_6_universal",
        "macosx_10_5_x86_64",
        "macosx_10_5_intel",
        "macosx_10_5_fat64",
        "macosx_10_5_fat32",
        "macosx_10_5_universal2",
        "macosx_10_5_universal",
        "macosx_10_4_x86_64",
        "macosx_10_4_intel",
        "macosx_10_4_fat64",
        "macosx_10_4_fat32",
        "macosx_10_4_universal2",
        "macosx_10_4_universal",
    ]

    tags = Platform(os.Macos(14, 0), Arch.X86_64).compatible_tags
    assert tags == [
        "macosx_14_0_x86_64",
        "macosx_14_0_intel",
        "macosx_14_0_fat64",
        "macosx_14_0_fat32",
        "macosx_14_0_universal2",
        "macosx_14_0_universal",
        "macosx_13_0_x86_64",
        "macosx_13_0_intel",
        "macosx_13_0_fat64",
        "macosx_13_0_fat32",
        "macosx_13_0_universal2",
        "macosx_13_0_universal",
        "macosx_12_0_x86_64",
        "macosx_12_0_intel",
        "macosx_12_0_fat64",
        "macosx_12_0_fat32",
        "macosx_12_0_universal2",
        "macosx_12_0_universal",
        "macosx_11_0_x86_64",
        "macosx_11_0_intel",
        "macosx_11_0_fat64",
        "macosx_11_0_fat32",
        "macosx_11_0_universal2",
        "macosx_11_0_universal",
        "macosx_10_16_x86_64",
        "macosx_10_16_intel",
        "macosx_10_16_fat64",
        "macosx_10_16_fat32",
        "macosx_10_16_universal2",
        "macosx_10_16_universal",
        "macosx_10_15_x86_64",
        "macosx_10_15_intel",
        "macosx_10_15_fat64",
        "macosx_10_15_fat32",
        "macosx_10_15_universal2",
        "macosx_10_15_universal",
        "macosx_10_14_x86_64",
        "macosx_10_14_intel",
        "macosx_10_14_fat64",
        "macosx_10_14_fat32",
        "macosx_10_14_universal2",
        "macosx_10_14_universal",
        "macosx_10_13_x86_64",
        "macosx_10_13_intel",
        "macosx_10_13_fat64",
        "macosx_10_13_fat32",
        "macosx_10_13_universal2",
        "macosx_10_13_universal",
        "macosx_10_12_x86_64",
        "macosx_10_12_intel",
        "macosx_10_12_fat64",
        "macosx_10_12_fat32",
        "macosx_10_12_universal2",
        "macosx_10_12_universal",
        "macosx_10_11_x86_64",
        "macosx_10_11_intel",
        "macosx_10_11_fat64",
        "macosx_10_11_fat32",
        "macosx_10_11_universal2",
        "macosx_10_11_universal",
        "macosx_10_10_x86_64",
        "macosx_10_10_intel",
        "macosx_10_10_fat64",
        "macosx_10_10_fat32",
        "macosx_10_10_universal2",
        "macosx_10_10_universal",
        "macosx_10_9_x86_64",
        "macosx_10_9_intel",
        "macosx_10_9_fat64",
        "macosx_10_9_fat32",
        "macosx_10_9_universal2",
        "macosx_10_9_universal",
        "macosx_10_8_x86_64",
        "macosx_10_8_intel",
        "macosx_10_8_fat64",
        "macosx_10_8_fat32",
        "macosx_10_8_universal2",
        "macosx_10_8_universal",
        "macosx_10_7_x86_64",
        "macosx_10_7_intel",
        "macosx_10_7_fat64",
        "macosx_10_7_fat32",
        "macosx_10_7_universal2",
        "macosx_10_7_universal",
        "macosx_10_6_x86_64",
        "macosx_10_6_intel",
        "macosx_10_6_fat64",
        "macosx_10_6_fat32",
        "macosx_10_6_universal2",
        "macosx_10_6_universal",
        "macosx_10_5_x86_64",
        "macosx_10_5_intel",
        "macosx_10_5_fat64",
        "macosx_10_5_fat32",
        "macosx_10_5_universal2",
        "macosx_10_5_universal",
        "macosx_10_4_x86_64",
        "macosx_10_4_intel",
        "macosx_10_4_fat64",
        "macosx_10_4_fat32",
        "macosx_10_4_universal2",
        "macosx_10_4_universal",
    ]

    tags = Platform(os.Macos(10, 6), Arch.X86_64).compatible_tags
    assert tags == [
        "macosx_10_6_x86_64",
        "macosx_10_6_intel",
        "macosx_10_6_fat64",
        "macosx_10_6_fat32",
        "macosx_10_6_universal2",
        "macosx_10_6_universal",
        "macosx_10_5_x86_64",
        "macosx_10_5_intel",
        "macosx_10_5_fat64",
        "macosx_10_5_fat32",
        "macosx_10_5_universal2",
        "macosx_10_5_universal",
        "macosx_10_4_x86_64",
        "macosx_10_4_intel",
        "macosx_10_4_fat64",
        "macosx_10_4_fat32",
        "macosx_10_4_universal2",
        "macosx_10_4_universal",
    ]


def test_platform_tags_windows():
    tags = Platform(os.Windows(), Arch.X86_64).compatible_tags
    assert tags == ["win_amd64"]


def test_platform_tags_musl():
    tags = Platform(os.Musllinux(1, 2), Arch.Aarch64).compatible_tags
    assert tags == ["linux_aarch64", "musllinux_1_1_aarch64", "musllinux_1_2_aarch64"]


@pytest.mark.parametrize(
    "text,expected,normalized",
    [
        ("linux", Platform(os.Manylinux(2, 40), Arch.X86_64), "manylinux_2_40_x86_64"),
        ("macos", Platform(os.Macos(14, 0), Arch.Aarch64), "macos_14_0_arm64"),
        ("windows", Platform(os.Windows(), Arch.X86_64), "windows_amd64"),
        ("alpine", Platform(os.Musllinux(1, 2), Arch.X86_64), "musllinux_1_2_x86_64"),
        (
            "manylinux_2_20_aarch64",
            Platform(os.Manylinux(2, 20), Arch.Aarch64),
            "manylinux_2_20_aarch64",
        ),
        (
            "macos_14_0_arm64",
            Platform(os.Macos(14, 0), Arch.Aarch64),
            "macos_14_0_arm64",
        ),
        ("windows_amd64", Platform(os.Windows(), Arch.X86_64), "windows_amd64"),
        ("windows_arm64", Platform(os.Windows(), Arch.Aarch64), "windows_arm64"),
        (
            "macos_12_0_x86_64",
            Platform(os.Macos(12, 0), Arch.X86_64),
            "macos_12_0_x86_64",
        ),
        (
            "mingw_x86_64",
            Platform(os.Generic("mingw"), Arch.X86_64),
            "mingw_x86_64",
        ),
    ],
)
def test_parse_platform(text, expected, normalized):
    platform = Platform.parse(text)
    assert platform == expected
    assert str(platform) == normalized



from __future__ import annotations

import abc
import typing as t

from packaging.specifiers import SpecifierSet
from packaging.version import Version

UnparsedVersion = t.Union[Version, str]


class InvalidSpecifier(ValueError):
    pass


class BaseSpecifier(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __str__(self) -> str:
        """
        Returns the str representation of this Specifier-like object. This
        should be representative of the Specifier itself.
        """

    @abc.abstractmethod
    def __and__(self, other: t.Any) -> BaseSpecifier:
        raise NotImplementedError

    @abc.abstractmethod
    def __or__(self, other: t.Any) -> BaseSpecifier:
        raise NotImplementedError

    @abc.abstractmethod
    def __invert__(self) -> BaseSpecifier:
        raise NotImplementedError

    def is_simple(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self}>"

    def is_empty(self) -> bool:
        return False

    def is_any(self) -> bool:
        return False

    @abc.abstractmethod
    def __contains__(self, value: str) -> bool:
        raise NotImplementedError


class VersionSpecifier(BaseSpecifier):
    @abc.abstractmethod
    def contains(
        self, version: UnparsedVersion, prereleases: bool | None = None
    ) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def num_parts(self) -> int:
        raise NotImplementedError

    def __contains__(self, version: UnparsedVersion) -> bool:
        return self.contains(version)

    @abc.abstractmethod
    def to_specifierset(self) -> SpecifierSet:
        """Convert to a packaging.specifiers.SpecifierSet object."""
        raise NotImplementedError



# Adapted from poetry/core/version/markers.py
# The original work is published under the MIT license.
# Copyright (c) 2020 Sébastien Eustace
# Adapted by Frost Ming (c) 2023

from __future__ import annotations

import functools
from typing import TYPE_CHECKING

from packaging.markers import InvalidMarker as _InvalidMarker
from packaging.markers import Marker as _Marker

from dep_logic.markers.any import AnyMarker
from dep_logic.markers.base import BaseMarker
from dep_logic.markers.empty import EmptyMarker
from dep_logic.markers.multi import MultiMarker
from dep_logic.markers.single import MarkerExpression
from dep_logic.markers.union import MarkerUnion
from dep_logic.utils import get_reflect_op

if TYPE_CHECKING:
    from typing import List, Literal, Tuple, Union

    from packaging.markers import Op, Value, Variable

    _ParsedMarker = Tuple[Variable, Op, Value]
    _ParsedMarkers = Union[
        _ParsedMarker, List[Union["_ParsedMarkers", Literal["or", "and"]]]
    ]


__all__ = [
    "AnyMarker",
    "BaseMarker",
    "EmptyMarker",
    "InvalidMarker",
    "MarkerExpression",
    "MarkerUnion",
    "MultiMarker",
    "from_pkg_marker",
    "parse_marker",
]


class InvalidMarker(ValueError):
    """
    An invalid marker was found, users should refer to PEP 508.
    """


@functools.lru_cache(maxsize=None)
def parse_marker(marker: str) -> BaseMarker:
    if marker == "<empty>":
        return EmptyMarker()

    if not marker or marker == "*":
        return AnyMarker()
    try:
        parsed = _Marker(marker)
    except _InvalidMarker as e:
        raise InvalidMarker(str(e)) from e

    markers = _build_markers(parsed._markers)

    return markers


def from_pkg_marker(marker: _Marker) -> BaseMarker:
    return _build_markers(marker._markers)


def _build_markers(markers: _ParsedMarkers) -> BaseMarker:
    from packaging.markers import Variable

    if isinstance(markers, tuple):
        if isinstance(markers[0], Variable):
            name, op, value, reversed = (
                str(markers[0]),
                str(markers[1]),
                str(markers[2]),
                False,
            )
        else:
            # in reverse order
            name, op, value, reversed = (
                str(markers[2]),
                get_reflect_op(str(markers[1])),
                str(markers[0]),
                True,
            )
        return MarkerExpression(name, op, value, reversed)
    or_groups: list[BaseMarker] = [AnyMarker()]
    for item in markers:
        if item == "or":
            or_groups.append(AnyMarker())
        elif item == "and":
            continue
        else:
            or_groups[-1] &= _build_markers(item)
    return MarkerUnion.of(*or_groups)


def _patch_marker_parser() -> None:
    import re

    try:
        from packaging._tokenizer import DEFAULT_RULES
    except (ModuleNotFoundError, AttributeError):
        return

    DEFAULT_RULES["VARIABLE"] = re.compile(
        r"""
            \b(
                python_version
                |python_full_version
                |os[._]name
                |sys[._]platform
                |platform_(release|system)
                |platform[._](version|machine|python_implementation)
                |python_implementation
                |implementation_(name|version)
                |extras?
                |dependency_groups
            )\b
        """,
        re.VERBOSE,
    )


_patch_marker_parser()
del _patch_marker_parser



import pytest

from dep_logic.tags import EnvSpec
from dep_logic.tags.tags import EnvCompatibility


def test_check_wheel_tags():
    wheels = [
        "protobuf-5.27.2-cp313-cp313t-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-cp313-cp313-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-cp310-abi3-win32.whl",
        "protobuf-5.27.2-cp310-abi3-win_amd64.whl",
        "protobuf-5.27.2-cp310-cp310-macosx_12_0_arm64.whl",
        "protobuf-5.27.2-cp38-abi3-macosx_10_9_universal2.whl",
        "protobuf-5.27.2-cp38-abi3-manylinux2014_aarch64.whl",
        "protobuf-5.27.2-cp38-abi3-manylinux2014_x86_64.whl",
        "protobuf-5.27.2-cp38-cp38-win32.whl",
        "protobuf-5.27.2-cp38-cp38-win_amd64.whl",
        "protobuf-5.27.2-cp39-cp39-win32.whl",
        "protobuf-5.27.2-cp39-cp39-win_amd64.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]

    linux_env = EnvSpec.from_spec(">=3.9", "linux", "cpython")
    wheel_compats = {
        f: c
        for f, c in {f: linux_env.wheel_compatibility(f) for f in wheels}.items()
        if c is not None
    }
    filtered_wheels = sorted(wheel_compats, key=wheel_compats.__getitem__, reverse=True)
    assert filtered_wheels == [
        "protobuf-5.27.2-cp38-abi3-manylinux2014_x86_64.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]

    windows_env = EnvSpec.from_spec(">=3.9", "windows", "cpython")
    wheel_compats = {
        f: c
        for f, c in {f: windows_env.wheel_compatibility(f) for f in wheels}.items()
        if c is not None
    }
    filtered_wheels = sorted(wheel_compats, key=wheel_compats.__getitem__, reverse=True)
    assert filtered_wheels == [
        "protobuf-5.27.2-cp310-abi3-win_amd64.whl",
        "protobuf-5.27.2-cp39-cp39-win_amd64.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]

    macos_env = EnvSpec.from_spec(">=3.9", "macos", "cpython")
    wheel_compats = {
        f: c
        for f, c in {f: macos_env.wheel_compatibility(f) for f in wheels}.items()
        if c is not None
    }
    filtered_wheels = sorted(wheel_compats, key=wheel_compats.__getitem__, reverse=True)
    assert filtered_wheels == [
        "protobuf-5.27.2-cp313-cp313-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-cp310-cp310-macosx_12_0_arm64.whl",
        "protobuf-5.27.2-cp38-abi3-macosx_10_9_universal2.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]

    macos_free_threaded_env = EnvSpec.from_spec(">=3.9", "macos", "cpython", True)
    wheel_compats = {
        f: c
        for f, c in {
            f: macos_free_threaded_env.wheel_compatibility(f) for f in wheels
        }.items()
        if c is not None
    }
    filtered_wheels = sorted(wheel_compats, key=wheel_compats.__getitem__, reverse=True)
    assert filtered_wheels == [
        "protobuf-5.27.2-cp313-cp313t-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]

    python_env = EnvSpec.from_spec(">=3.9")
    wheel_compats = {
        f: c
        for f, c in {f: python_env.wheel_compatibility(f) for f in wheels}.items()
        if c is not None
    }
    filtered_wheels = sorted(wheel_compats, key=wheel_compats.__getitem__, reverse=True)
    assert filtered_wheels == [
        "protobuf-5.27.2-cp313-cp313t-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-cp313-cp313-macosx_14_0_arm64.whl",
        "protobuf-5.27.2-cp310-cp310-macosx_12_0_arm64.whl",
        "protobuf-5.27.2-cp310-abi3-win32.whl",
        "protobuf-5.27.2-cp310-abi3-win_amd64.whl",
        "protobuf-5.27.2-cp39-cp39-win32.whl",
        "protobuf-5.27.2-cp39-cp39-win_amd64.whl",
        "protobuf-5.27.2-cp38-abi3-macosx_10_9_universal2.whl",
        "protobuf-5.27.2-cp38-abi3-manylinux2014_aarch64.whl",
        "protobuf-5.27.2-cp38-abi3-manylinux2014_x86_64.whl",
        "protobuf-5.27.2-py3-none-any.whl",
    ]


@pytest.mark.parametrize(
    "left,right,expected",
    [
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec(">=3.9", "macos"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos"),
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec(">=3.7,<3.10"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
        (
            EnvSpec.from_spec(">=3.7,<3.10"),
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec("<3.8", "macos", "cpython"),
            EnvCompatibility.INCOMPATIBLE,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec("<3.10", "linux", "cpython"),
            EnvCompatibility.INCOMPATIBLE,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos", "cpython"),
            EnvSpec.from_spec("<3.10", "macos", "pypy"),
            EnvCompatibility.INCOMPATIBLE,
        ),
        (
            EnvSpec.from_spec(">=3.9", "macos_x86_64", "cpython"),
            EnvSpec.from_spec("<3.10", "macos_10_9_x86_64", "cpython"),
            EnvCompatibility.HIGHER,
        ),
        (
            EnvSpec.from_spec("<3.10", "macos_10_9_x86_64", "cpython"),
            EnvSpec.from_spec(">=3.9", "macos_x86_64", "cpython"),
            EnvCompatibility.LOWER_OR_EQUAL,
        ),
    ],
)
def test_env_spec_comparison(left, right, expected):
    assert left.compare(right) == expected
